from typing import Any, Optional, Protocol

MISSING = object()

class CacheBackend(Protocol):
    """Protocol/Interface para backends de persistência."""
    
    def get(self, key: str) -> Optional[Any]: ...
    def set(self, key: str, value: Any, ttl: Optional[float] = None) -> None: ...
    def delete(self, key: str) -> None: ...
    def clear(self) -> None: ...
    def cleanup_expired(self) -> int: ...


import time
import pickle
import hashlib
import json
from threading import RLock
from typing import Any, Optional
from pathlib import Path

from .cache_backend import CacheBackend, MISSING


class FileSystemBackend(CacheBackend):
    """Backend de persistência em filesystem com auto-limpeza."""
    
    def __init__(
        self,
        cache_dir: str = ".cache",
        file_ttl_days: float = 7,
        auto_cleanup: bool = True
    ):
        self.cache_dir = Path(cache_dir)
        self.file_ttl_seconds = file_ttl_days * 86400
        self.auto_cleanup = auto_cleanup
        self._lock = RLock()
        
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.metadata_file = self.cache_dir / ".cache_metadata.json"
        self._metadata = self._load_metadata()
        
        if self.auto_cleanup:
            self.cleanup_expired()
    
    def _load_metadata(self) -> dict:
        """Carrega metadata de expiração."""
        if not self.metadata_file.exists():
            return {}
        
        try:
            with open(self.metadata_file, 'r') as f:
                return json.load(f)
        except (json.JSONDecodeError, IOError):
            return {}
    
    def _save_metadata(self) -> None:
        """Salva metadata de expiração."""
        try:
            with open(self.metadata_file, 'w') as f:
                json.dump(self._metadata, f)
        except IOError:
            pass  # Non-critical failure
    
    def _get_file_path(self, key: str) -> Path:
        """Gera caminho do arquivo baseado na chave."""
        key_hash = hashlib.sha256(key.encode()).hexdigest()
        return self.cache_dir / f"cache_{key_hash}.pkl"
    
    def _is_expired(self, file_path: Path) -> bool:
        """Verifica se arquivo está expirado."""
        key = file_path.name
        
        # Check metadata first
        if key in self._metadata:
            return time.time() > self._metadata[key]
        
        # Fallback to file timestamp
        try:
            mtime = file_path.stat().st_mtime
            return (time.time() - mtime) > self.file_ttl_seconds
        except OSError:
            return True
    
    def get(self, key: str) -> Optional[Any]:
        """Recupera valor do filesystem."""
        with self._lock:
            file_path = self._get_file_path(key)
            
            if not file_path.exists() or self._is_expired(file_path):
                if file_path.exists():
                    self.delete(key)
                return MISSING
            
            try:
                with open(file_path, 'rb') as f:
                    return pickle.load(f)
            except (pickle.PickleError, IOError, EOFError):
                self.delete(key)
                return MISSING
    
    def set(self, key: str, value: Any, ttl: Optional[float] = None) -> None:
        """Armazena valor no filesystem."""
        with self._lock:
            file_path = self._get_file_path(key)
            
            try:
                with open(file_path, 'wb') as f:
                    pickle.dump(value, f, protocol=pickle.HIGHEST_PROTOCOL)
                
                expiry_time = time.time() + (ttl or self.file_ttl_seconds)
                self._metadata[file_path.name] = expiry_time
                self._save_metadata()
            except (pickle.PickleError, IOError):
                if file_path.exists():
                    file_path.unlink()
    
    def delete(self, key: str) -> None:
        """Remove valor do filesystem."""
        with self._lock:
            file_path = self._get_file_path(key)
            
            if file_path.exists():
                try:
                    file_path.unlink()
                except OSError:
                    pass
            
            if file_path.name in self._metadata:
                del self._metadata[file_path.name]
                self._save_metadata()
    
    def clear(self) -> None:
        """Limpa todos os arquivos de cache."""
        with self._lock:
            for cache_file in self.cache_dir.glob("cache_*.pkl"):
                try:
                    cache_file.unlink()
                except OSError:
                    pass
            
            self._metadata.clear()
            self._save_metadata()
    
    def cleanup_expired(self) -> int:
        """Remove arquivos expirados. Retorna número de itens removidos."""
        with self._lock:
            removed_count = 0
            
            for cache_file in self.cache_dir.glob("cache_*.pkl"):
                if self._is_expired(cache_file):
                    try:
                        cache_file.unlink()
                        removed_count += 1
                        
                        if cache_file.name in self._metadata:
                            del self._metadata[cache_file.name]
                    except OSError:
                        pass
            
            if removed_count > 0:
                self._save_metadata()
            
            return removed_count
    
    def get_stats(self) -> dict:
        """Retorna estatísticas do cache em disco."""
        with self._lock:
            cache_files   = list(self.cache_dir.glob("cache_*.pkl"))
            total_size    = sum(f.stat().st_size for f in cache_files if f.exists())
            expired_count = sum(1 for f in cache_files if self._is_expired(f))
            
            return {
                'total_files'      : len(cache_files),
                'expired_files'    : expired_count,
                'total_size_bytes' : total_size,
                'total_size_mb'    : round(total_size / (1024 * 1024), 2),
                'cache_dir'        : str(self.cache_dir.absolute())
            }




from typing import Any
import re


class KeyPathResolver:
    """Resolve caminhos aninhados em objetos (ex: 'user.addresses[0].city')."""
    
    INDEX_PATTERN = re.compile(r'^(.+?)\[(\d+)\]$')
    
    @classmethod
    def resolve(cls, obj: Any, path: str) -> Any:
        """Resolve um caminho aninhado em um objeto."""
        if not path:
            return obj
        
        current = obj
        for part in path.split('.'):
            current = cls._resolve_part(current, part, path)
        
        return current
    
    @classmethod
    def _resolve_part(cls, obj: Any, part: str, full_path: str) -> Any:
        """Resolve uma parte do caminho (atributo ou índice)."""
        match = cls.INDEX_PATTERN.match(part)
        
        if match:
            attr_name, index = match.groups()
            obj = cls._access_attr_or_key(obj, attr_name)
            return cls._access_index(obj, int(index), full_path)
        
        return cls._access_attr_or_key(obj, part)
    
    @staticmethod
    def _access_attr_or_key(obj: Any, key: str) -> Any:
        """Acessa atributo ou chave de dict."""
        # Check dict/list access FIRST
        if hasattr(obj, '__getitem__'):
            try:
                return obj[key]  # Try as dict key or list index first
            except (KeyError, TypeError, IndexError):
                pass  # Fall through to attribute access
        
        # Then check attributes
        if hasattr(obj, key):
            return getattr(obj, key)
        
        raise AttributeError(f"Cannot access '{key}' on {type(obj).__name__}")
    
    @staticmethod
    def _access_index(obj: Any, index: int, path: str) -> Any:
        """Acessa índice em sequência."""
        try:
            return obj[index]
        except (IndexError, KeyError, TypeError) as e:
            raise KeyError(f"Cannot access index [{index}] in path '{path}': {e}") from e




from .cache_backend import CacheBackend
from .cache_entry import CacheEntry
from .cache_key_builder import CacheKeyBuilder
from .cache_stats import CacheStats
from .decorator import remember
from .filesystem_backend import FileSystemBackend
from .key_path_resolver import KeyPathResolver
from .lru_cache import LRUCache, MISSING

__all__ = [
    "CacheBackend",
    "CacheEntry",
    "CacheKeyBuilder",
    "CacheStats",
    "remember",
    "FileSystemBackend",
    "KeyPathResolver",
    "LRUCache",
    "MISSING"
]


import hashlib
from collections import OrderedDict
from threading import RLock
from typing import Any, Optional, Tuple

from .cache_backend import CacheBackend, MISSING
from .cache_entry import CacheEntry


class LRUCache:
    """Cache LRU thread-safe com suporte a TTL e persistência."""
    
    def __init__(
        self,
        capacity: int,
        ttl: Optional[float] = None,
        backend: Optional[CacheBackend] = None
    ):
        if capacity <= 0:
            raise ValueError("Capacity must be positive")
        
        self.cache    = OrderedDict()
        self.capacity = capacity
        self.ttl      = ttl
        self.backend  = backend
        self._lock    = RLock()

    def get(self, key: Tuple) -> Optional[Any]:
        """Obtém valor do cache (memória + backend)."""
        with self._lock:
            # Try memory first
            value = self._get_from_memory(key)
            if value is not MISSING:
                return value
            
            # Fallback to backend
            return self._get_from_backend(key)
    
    def _get_from_memory(self, key: Tuple) -> Optional[Any]:
        """Tenta recuperar da memória."""
        if key not in self.cache:
            return MISSING
        
        entry = self.cache[key]
        
        if entry.is_expired(self.ttl):
            del self.cache[key]
            return MISSING
        
        self.cache.move_to_end(key)
        return entry.value
    
    def _get_from_backend(self, key: Tuple) -> Optional[Any]:
        """Tenta recuperar do backend."""
        if not self.backend:
            return MISSING
        
        try:
            key_str = self._serialize_key(key)
            value   = self.backend.get(key_str)
            
            if value is not MISSING:
                # Reconstruct in memory
                self.cache[key] = CacheEntry(value)
                self.cache.move_to_end(key)
                return value
        except Exception:
            pass  # Backend failures shouldn't crash
        
        return MISSING

    def put(self, key: Tuple, value: Any) -> None:
        """Adiciona valor no cache (memória + backend)."""
        with self._lock:
             self._put_in_memory(key, value)
             self._put_in_backend(key, value)
    
    def _put_in_memory(self, key: Tuple, value: Any) -> None:
        """Armazena na memória."""
        self.cache[key] = CacheEntry(value)
        self.cache.move_to_end(key)
        
        if len(self.cache) > self.capacity:
            self.cache.popitem(last=False)
    
    def _put_in_backend(self, key: Tuple, value: Any) -> None:
        """Armazena no backend."""
        if not self.backend:
            return
        
        try:
            key_str = self._serialize_key(key)
            self.backend.set(key_str, value, ttl=self.ttl)
        except Exception:
            pass  # Backend failures shouldn't crash
    
    def clear(self) -> None:
        """Limpa cache de memória."""
        with self._lock:
             self.cache.clear()
    
    def size(self) -> int:
        """Retorna tamanho atual do cache em memória."""
        with self._lock:
            return len(self.cache)
    
    @staticmethod
    def _serialize_key(key: Tuple) -> str:
        """Serializa chave tupla para string."""
        return hashlib.md5(str(key).encode()).hexdigest()



from typing import Any, Callable, Optional, Tuple, List

from .key_path_resolver import KeyPathResolver


class CacheKeyBuilder:
    """Construtor de chaves para cache com suporte a paths aninhados."""
    
    def __init__(
        self,
        func: Callable,
        key_args: Optional[List[str]],
        typed: bool,
        use_class_name: bool
    ):
        self.func = func
        self.key_args = key_args
        self.typed = typed
        self.use_class_name = use_class_name
        self._param_names = func.__code__.co_varnames[:func.__code__.co_argcount]

    def build(self, self_obj: Any, args: Tuple, kwargs: dict) -> Tuple:
        """Constrói chave de cache baseada nos argumentos."""
        try:
            if self.key_args is None:
                key = self._build_full_key(self_obj, args, kwargs)
            else:
                key = self._build_selective_key(self_obj, args, kwargs)
            
            if self.typed:
                key = self._add_type_info(key, args, kwargs)
            
            hash(key)  # Ensure hashable
            return key
            
        except TypeError as e:
            raise TypeError(f"Cache key contains unhashable type: {e}") from e

    def _build_full_key(self, self_obj: Any, args: Tuple, kwargs: dict) -> Tuple:
        """Constrói chave usando todos os argumentos."""
        self_key = self._get_self_key(self_obj)
        kwargs_tuple = tuple(sorted(kwargs.items())) if kwargs else ()
        return (self_key,) + args + kwargs_tuple

    def _build_selective_key(self, self_obj: Any, args: Tuple, kwargs: dict) -> Tuple:
        """Constrói chave usando apenas argumentos especificados."""
        args_dict = dict(zip(self._param_names[1:], args))
        args_dict.update(kwargs)
        
        self_key = self._get_self_key(self_obj)
        key_parts = [self_key]
        
        assert self.key_args is not None
        for arg_path in self.key_args:
            value = self._resolve_arg_path(arg_path, args_dict)
            key_parts.append(value)
        
        return tuple(key_parts)
    
    def _resolve_arg_path(self, arg_path: str, args_dict: dict) -> Any:
        """Resolve um path de argumento (simples ou aninhado)."""
        if '.' not in arg_path and '[' not in arg_path:
            # Simple argument
            if arg_path not in args_dict:
                raise KeyError(f"Argument '{arg_path}' not found")
            return args_dict[arg_path]
        
        # Nested path
        base_arg = arg_path.split('.')[0].split('[')[0]
        
        if base_arg not in args_dict:
            raise KeyError(f"Base argument '{base_arg}' not found")
        
        try:
            remaining_path = arg_path[len(base_arg):].lstrip('.')
            return KeyPathResolver.resolve(args_dict[base_arg], remaining_path)
        except (AttributeError, KeyError, IndexError) as e:
            raise KeyError(f"Cannot resolve path '{arg_path}': {e}") from e

    def _add_type_info(self, key: Tuple, args: Tuple, kwargs: dict) -> Tuple:
        """Adiciona informação de tipo à chave."""
        type_info = (
            tuple(type(arg).__name__ for arg in args) +
            tuple(type(v).__name__ for k, v in sorted(kwargs.items()))
        )
        return key + type_info
    
    def _get_self_key(self, self_obj: Any) -> str:
        """Gera chave estável ou por instância para self."""
        if self.use_class_name:
            return f"{self_obj.__class__.__module__}.{self_obj.__class__.__name__}"
        return str(id(self_obj))




import time
from typing import Any, Optional


class CacheEntry:
    """Entrada individual do cache com timestamp."""
    
    __slots__ = ('value', 'timestamp')
    
    def __init__(self, value: Any):
        self.value = value
        self.timestamp = time.time()
    
    def is_expired(self, ttl: Optional[float]) -> bool:
        """Verifica se a entrada expirou."""
        return ttl is not None and (time.time() - self.timestamp) > ttl


import asyncio
from typing import Any, Callable, Optional, List, TypeVar

from .func_wrapper import create_wrapper
from .cache_backend import CacheBackend
from .filesystem_backend import FileSystemBackend
from .lru_cache import LRUCache
from .cache_key_builder import CacheKeyBuilder
from .cache_stats import CacheStats


F = TypeVar('F', bound=Callable[..., Any])

def remember(
    maxsize: int = 128,
    typed: bool = False,
    key_args: Optional[List[str]] = None,
    ttl_seconds: Optional[float] = 300,
    persist: bool = False,
    backend: Optional[CacheBackend] = None,
    cache_dir: str = ".cache",
    file_ttl_days: int = 1,
    per_instance: bool = False
) -> Callable[[F], F]:
    """
    Decorador de cache LRU thread-safe com persistência opcional e compatível com async.
    
    Args:
        maxsize       : Número máximo de itens no cache em memória
        typed         : Se True, tipos diferentes são cacheados separadamente
        key_args      : Lista de argumentos/paths para chave (None = todos)
        ttl_seconds   : TTL para entradas em memória (None = sem expiração)
        persist       : Se True, habilita persistência em disco
        backend       : Backend customizado (None = usa FileSystemBackend)
        cache_dir     : Diretório para cache (usado se persist=True e backend=None)
        file_ttl_days : Dias até arquivos expirarem (padrão: 1)
        per_instance  : Se True, cache separado por instância (usa id()).
                        Se False e persist=True, cache compartilhado entre execuções
    
    Returns:
        Função decorada com cache (compatível com sync e async)
    """
    if maxsize <= 0:
        raise ValueError("maxsize must be positive")
    if ttl_seconds is not None and ttl_seconds <= 0:
        raise ValueError("ttl_seconds must be positive or None")
    if file_ttl_days <= 0:
        raise ValueError("file_ttl_days must be positive")
    
    def decorator(func: F) -> F:
        # Setup backend
        cache_backend = None
        is_persistent = persist or (backend is not None)
        # persist=True ativa o file backend por padrão
        # porém se um backend customizado for fornecido,
        # fica implicito que o usuário quer usá-lo
        if is_persistent:
            cache_backend = backend or FileSystemBackend(
                cache_dir     = cache_dir,
                file_ttl_days = file_ttl_days,
                auto_cleanup  = True
            )
        
        # Initialize components
        cache = LRUCache(maxsize, ttl=ttl_seconds, backend=cache_backend)
        stats = CacheStats()
        
        use_stable_key = is_persistent and not per_instance
        key_builder = CacheKeyBuilder(func, key_args, typed, use_class_name=use_stable_key)
        
        # Detect if function is async
        is_async = asyncio.iscoroutinefunction(func)
        
        # Create appropriate wrapper
        wrapper = create_wrapper(func, cache, stats, key_builder, is_async)

        
        def cache_info():
            """Retorna informações sobre o cache."""
            hits, misses = stats.get_stats()
            info = {
                'hits'         : hits,
                'misses'       : misses,
                'currsize'     : cache.size(),
                'hit_rate'     : stats.hit_rate(),
                'is_async'     : is_async,
                'maxsize'      : maxsize,
                'per_instance' : per_instance,
                'persistent'   : persist or (cache_backend is not None),
            }
            
            if cache_backend and hasattr(cache_backend, 'get_stats'):
                info['backend_stats'] = cache_backend.get_stats() # type: ignore
            
            return info
        
        wrapper.cache_info    = cache_info     # type: ignore
        wrapper.stats         = stats          # type: ignore
        wrapper.cache         = cache          # type: ignore
        
        return wrapper  # type: ignore
    
    return decorator


from threading import RLock
from typing import Tuple


class CacheStats:
    """Estatísticas de uso do cache thread-safe."""
    
    __slots__ = ('hits', 'misses', '_lock')
    
    def __init__(self):
        self.hits   = 0
        self.misses = 0
        self._lock  = RLock()
    
    def record_hit(self):
        with self._lock:
             self.hits += 1
    
    def record_miss(self):
        with self._lock:
             self.misses += 1
    
    def get_stats(self) -> Tuple[int, int]:
        with self._lock:
            return (self.hits, self.misses)
    
    def hit_rate(self) -> float:
        with self._lock:
            total = self.hits + self.misses
            return round(self.hits / total, 4) if total > 0 else 0.0




from functools import wraps
from typing import Callable

from .lru_cache import LRUCache
from .cache_key_builder import CacheKeyBuilder
from .cache_stats import CacheStats
from .cache_backend import MISSING


def create_wrapper(
    func       : Callable,
    cache      : LRUCache,
    stats      : CacheStats,
    key_builder: CacheKeyBuilder,
    is_async   : bool
) -> Callable:
    """Cria wrapper apropriado (sync ou async)."""
    
    if is_async:
        @wraps(func)
        async def async_wrapper(self, *args, **kwargs):
            key = key_builder.build(self, args, kwargs)
            cached_value = cache.get(key)
            if cached_value is not MISSING:
                stats.record_hit()
                return cached_value
            
            stats.record_miss()
            result = await func(self, *args, **kwargs)
            cache.put(key, result)
            
            return result
        
        return async_wrapper
    
    else:
        @wraps(func)
        def sync_wrapper(self, *args, **kwargs):
            key = key_builder.build(self, args, kwargs)
            cached_value = cache.get(key)
            if cached_value is not MISSING:
                stats.record_hit()
                return cached_value
            
            stats.record_miss()
            result = func(self, *args, **kwargs)
            cache.put(key, result)
            
            return result
        
        return sync_wrapper